import multiprocessing

DEFAULT_CHROMA_SQLITE_FILE = "chroma.sqlite3"
DEFAULT_TOPIC_NAMESPACE = "default"
DEFAULT_TENANT_ID = "default"
DEFAULT_DATABASE = "default_database"
DEFAULT_TENANT = "default_tenant"
DEFAULT_DISTANCE_METRIC = "l2"
DEFAULT_M = 16
DEFAULT_CONSTRUCTION_EF = 100
DEFAULT_SEARCH_EF = 10
DEFAULT_BATCH_SIZE = 100
DEFAULT_SYNC_THRESHOLD = 1000
DEFAULT_NUM_THREADS = multiprocessing.cpu_count()
DEFAULT_RESIZE_FACTOR = 1.2
DEFAULT_TOKENIZER = "trigram"
